#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.misc import safe_get
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.cfg import PlenopticamConfig
from plenopticam.lfp_aligner.cfa_processor import CfaProcessor
from plenopticam.lfp_reader.constants import SUPP_FILE_EXT
from plenopticam.lfp_reader.lfp_decoder import LfpDecoder

# external libs
import json
from os.path import join, exists, isdir, dirname, splitext, basename
from os import listdir
import tarfile


class CaliFinder(object):

    def __init__(self, cfg=None, sta=None):

        # input variables
        self.cfg = cfg if sta is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()

        # internal variables
        self._lfp_json = {}
        self._georef = None
        self._lensref = None
        self._serial = None
        self._cam_model = ''
        self._cal_fn = None
        self._raw_data = None
        self._file_found = None
        self._opt_prnt = True if sta is not None else self.cfg.params[self.cfg.opt_prnt]
        self._path = self.cfg.params[self.cfg.cal_path]

        # output variables
        self._wht_bay = None

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # auto calibration can only be used if calibration source path is either directory or tar archive
        if isdir(self._path) or self._path.lower().endswith('.tar'):

            # read JSON file from selected *.lfp image
            self._lfp_json = self.cfg.load_json(self.cfg.params[self.cfg.lfp_path])

            # extract serial number to support search
            self._serial = safe_get(self._lfp_json, 'camera', 'serialNumber')
            self._cam_model = self._serial if self._serial else safe_get(self._lfp_json, 'camera', 'model')

            # extract calibration reference data
            if self._cam_model.startswith(('A', 'F')):
                frames = safe_get(self._lfp_json, 'picture', 'derivationArray') #'frameArray')  #
                self._georef = frames[0]    #safe_get(frames[0], 'frame', 'imageRef') if frames else ''   #

            elif self._cam_model.startswith(('B', 'I')) or self._cam_model.isdigit():
                frames = safe_get(self._lfp_json, 'frames')
                self._georef = safe_get(frames[0], 'frame', 'geometryCorrectionRef') if frames else ''

            # print status
            if not self._serial and isdir(self._path):
                self.sta.status_msg('No serial number found in JSON file. Provide calibration file instead of folder',
                                    self._opt_prnt)
                self.sta.error = True

            # when path is directory
            if isdir(self._path):

                # look for geo data in calibration folders
                self._search_cal_dirs()

                # look for geo data in calibration tar-files (skip if already found in folders with file_found==True)
                self._search_cal_file()

            # when path is tar file
            elif self._path.lower().endswith('.tar'):

                # look for geo data in provided calibration tar-file
                self._search_tar_file(self._path)

            if not self._file_found and not self.sta.error:
                # print status and interrupt process
                self.sta.status_msg('White image file not found. Revise calibration path settings', self._opt_prnt)
                self.sta.error = True
            # load and keep white image if found and options are set or meta data is missing
            cond = self.cfg.params[self.cfg.opt_cali] or \
                   self.cfg.params[self.cfg.opt_vign] or \
                   not self.cfg.cond_meta_file()
            if self._file_found and cond:
                # convert raw data to image array and get metadata
                self._raw2img()

        return True

    def _raw2img(self):
        """ decode raw data to obtain bayer image and settings data """

        # skip if calibrated json file already exists, otherwise perform centroid calibration
        if self._raw_data:

            # decode raw data
            obj = LfpDecoder(self._raw_data, self.cfg, self.sta)
            obj.decode_raw()
            self._wht_bay = obj.bay_img
            del obj

            # balance Bayer channels in white image
            try:
                frame_arr = safe_get(self._wht_json, 'master', 'picture', 'frameArray')[0]
                self.cfg.lfpimg['ccm_wht'] = safe_get(frame_arr, 'frame', 'metadata', 'image', 'color', 'ccmRgbToSrgbArray')
                awb = safe_get(frame_arr, 'frame', 'metadata', 'devices', 'sensor', 'normalizedResponses')[0]
                gains = [1./awb['b'], 1./awb['r'], 1./awb['gr'], 1./awb['gb']]
                self.cfg.lfpimg['awb_wht'] = gains
            except ValueError:
                gains = [1/0.74476742744445801, 1/0.76306647062301636, 1, 1]

            # apply white balance gains to calibration file
            cfa_obj = CfaProcessor(bay_img=self._wht_bay, cfg=self.cfg, sta=self.sta)
            cfa_obj.set_gains(gains)
            self._wht_bay = cfa_obj.apply_awb()
            del cfa_obj

        return True

    def _match_lens(self, lens_spec):

        if lens_spec == self._lensref:
            self._cal_fn = lens_spec['name'].replace('.GCT', '.RAW')
            self._file_found = True

        return True

    def _match_georef(self, json_dict):
        """ compare georef hash value with that in provided json dictionary """

        # use JSON keys according to LFR type
        key1, key2 = ('calibrationFiles', 'hash') if self._cam_model.startswith(('B', 'I')) or self._cam_model.isdigit() else ('frame', 'imageRef')

        # search for georef hash value in geometry files of calibration folder
        for item in json_dict[key1]:
            if item[key2] == self._georef:  # stop when hash values match
                self._cal_fn = item['name'].replace('.GCT', '.RAW')     # save calibration file
                self._file_found = True
                break

        return True

    def _search_cal_dirs(self):
        """ look for geo data in calibration folders """

        # skip if file already found or if provided path is not a directory
        if not self._file_found:
            onlydirs = [d for d in listdir(self._path)]
            items = [self._serial] if onlydirs.count(self._serial) else onlydirs

            # iterate through directories
            for item in items:
                cali_manifest = join(join(self._path, item), 'cal_file_manifest.json')
                if exists(cali_manifest):
                    with open(cali_manifest, 'r') as file:
                        # find matching sha-1 value
                        json_dict = json.load(file)
                        self._match_georef(json_dict)
                        if self._file_found:
                            # update config and load raw data
                            self.cfg.params[self.cfg.cal_meta] = join(join(self._path.split('.')[0], self._serial), self._cal_fn)
                            with open(self.cfg.params[self.cfg.cal_meta], mode='rb') as meta_file:
                                self._raw_data = meta_file.read()
                            break

                # extract Lytro's 1st generation files
                elif item.lower().endswith(SUPP_FILE_EXT[-4:]):
                    fn = join(join(self._path, item))
                    if not exists(splitext(fn)[0]):
                        # status update
                        self.sta.status_msg('Extract ' + basename(fn), self.cfg.params[self.cfg.opt_prnt])
                        # bundle type decoding
                        with open(fn, mode='rb') as file:
                            obj = LfpDecoder(file, self.cfg, self.sta, lfp_path=fn)
                            obj.main()
                            del obj

                    # find matching file
                    mod_txts = [d for d in listdir(splitext(fn)[0])
                                if d.lower().startswith('mod') and d.lower().endswith('.txt')]
                    for mod_txt in mod_txts:
                        with open(join(splitext(fn)[0], mod_txt), 'r') as f:
                            json_dict = {}
                            json_dict.update(json.loads(f.read().rpartition('}')[0]+'}'))
                            #self._match_georef(json_dict['master']['picture']['frameArray'][0])
                            #if json_dict['master']['picture']['frameArray'][0]['frame']['imageRef'] == self._georef:
                            if json_dict['master']['picture']['derivationArray'] == self._georef:
                                self._file_found = True

        return True

    def _search_cal_file(self):
        """ look for geo data in calibration tar-files """

        # skip if file already found
        if not self._file_found:
            onlyfiles = [f for f in listdir(self._path) if f.lower().endswith('.tar')]
            tarstring = 'caldata-'+str(self._serial)+'.tar'
            fnames = [tarstring] if onlyfiles.count(tarstring) else onlyfiles

            # iterate through tar-files
            for fname in fnames:
                self._search_tar_file(join(self._path, fname))
                break

        return True

    def _search_tar_file(self, tarname):

        # verify tar file exists
        if not exists(tarname):
            self.sta.status_msg('Did not find tar-archive', opt=self._opt_prnt)
            self.sta.error = True
            return False

        # read mla_calibration JSON file from tar archive
        with tarfile.open(tarname, mode='r') as tar_obj:

            # find location of cal_file_manifest.json
            fname = 'cal_file_manifest.json'
            subdir = self._serial if self._serial + '/' + fname in tar_obj.getnames() else 'unitdata'
            fpath = subdir + '/' + fname    # join produces double backslash on Win causing errors in .extractfile()

            try:
                # extract cali file metadata
                cal_manifest = tar_obj.extractfile(fpath)
                json_dict = json.loads(cal_manifest.read().decode('utf-8'))
            except KeyError:
                self.sta.status_msg('Did not find "cal_file_manifest.json" in tar archive', opt=self._opt_prnt)
                self.sta.error = True
                return False

            # match hash value
            self._match_georef(json_dict)

            if self._cal_fn:
                self._file_found = True

                # update config
                self._serial = tarname.split('-')[-1].split('.')[0]
                tar_path = dirname(self._path) if self._path.lower().endswith('tar') else self._path
                self.cfg.params[self.cfg.cal_meta] = join(tar_path, self._serial,
                                                          self._cal_fn.lower().replace('.raw', '.json'))

                # load raw data (join produces double backslash on Win causing errors in .extractfile())
                self._raw_data = tar_obj.extractfile(subdir + '/' + self._cal_fn).read()
                json_file = tar_obj.extractfile(subdir + '/' + self._cal_fn.upper().replace('.RAW', '.TXT'))
                self._wht_json = json.loads(json_file.read())

    @property
    def raw_data(self):

        return self._raw_data

    @property
    def wht_bay(self):

        return self._wht_bay



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam import misc
from plenopticam.lfp_aligner.lfp_resampler import LfpResampler
from plenopticam.lfp_aligner.lfp_rotator import LfpRotator
from plenopticam.lfp_aligner.cfa_outliers import CfaOutliers
from plenopticam.lfp_aligner.cfa_processor import CfaProcessor
from plenopticam.lfp_aligner.lfp_devignetter import LfpDevignetter


class LfpAligner(object):

    def __init__(self, lfp_img, cfg=None, sta=None, wht_img=None):

        # input variables
        self.cfg = cfg
        self.sta = sta if sta is not None else misc.PlenopticamStatus()
        self._lfp_img = lfp_img.astype('float') if lfp_img is not None else None
        self._wht_img = wht_img.astype('float') if wht_img is not None else None

    def main(self):

        if self.cfg.lfpimg:
            # hot pixel correction
            obj = CfaOutliers(bay_img=self._lfp_img, cfg=self.cfg, sta=self.sta)
            obj.rectify_candidates_bayer(n=9, sig_lev=2.5)
            self._lfp_img = obj.bay_img
            del obj

        if self.cfg.params[self.cfg.opt_vign] and self._wht_img is not None:
            # apply de-vignetting
            obj = LfpDevignetter(lfp_img=self._lfp_img, wht_img=self._wht_img, cfg=self.cfg, sta=self.sta)
            obj.main()
            self._lfp_img = obj.lfp_img
            self._wht_img = obj.wht_img
            del obj

        if self.cfg.lfpimg and len(self._lfp_img.shape) == 2:
            # perform color filter array management and obtain rgb image
            cfa_obj = CfaProcessor(bay_img=self._lfp_img, wht_img=self._wht_img, cfg=self.cfg, sta=self.sta)
            cfa_obj.main()
            self._lfp_img = cfa_obj.rgb_img
            del cfa_obj

        if self.cfg.params[self.cfg.opt_rota] and self._lfp_img is not None:
            # de-rotate centroids
            obj = LfpRotator(self._lfp_img, self.cfg.calibs[self.cfg.mic_list], rad=None, cfg=self.cfg, sta=self.sta)
            obj.main()
            self._lfp_img, self.cfg.calibs[self.cfg.mic_list] = obj.lfp_img, obj.centroids
            del obj

        # interpolate each micro image with its MIC as the center with consistent micro image size
        obj = LfpResampler(lfp_img=self._lfp_img, cfg=self.cfg, sta=self.sta)
        obj.main()
        self._lfp_img = obj.lfp_img_align
        del obj

        return True

    @property
    def lfp_img(self):
        return self._lfp_img.copy()



from plenopticam.misc.data_proc import *
from plenopticam.misc.normalizer import Normalizer
from plenopticam.misc.os_ops import mkdir_p, rmdir_p, rm_file, select_file, get_img_list
from plenopticam.misc.file_rw import load_img_file, save_img_file, save_gif
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.misc.type_checks import *
from plenopticam.misc.errors import *
from plenopticam.misc.gamma_converter import GammaConverter
from plenopticam.misc.data_downloader import DataDownloader



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

import os
import sys
from tempfile import mkstemp

# local python files
from plenopticam import __version__
from plenopticam.gui.constants import PX, PY, ICON
from plenopticam.gui.widget_ctrl import CtrlWidget
from plenopticam.gui.widget_pbar import PbarWidget
from plenopticam.misc.status import PlenopticamStatus


class PlenopticamApp(tk.Tk):

    REL_PATH = os.path.join('icns', '1055104.gif')

    def __init__(self):

        # inheritance
        tk.Tk.__init__(self)

        # window title
        self.wm_title("PlenoptiCam-v"+__version__)

        # icon handling
        self.icon_handling()

        # initialize parameters
        self.sta = PlenopticamStatus()

        # instantiate controller
        self.ctrl_wid = CtrlWidget(self)
        self.ctrl_wid.pack(fill='both', expand=True, side='top', padx=PX, pady=PY)

        # instantiate view
        self.pbar_wid = PbarWidget(self)
        self.pbar_wid.pack(fill='both', expand=True, side='bottom', padx=PX, pady=PY)

    def icon_handling(self):
        """ load icon with OS temp folder if present or current working directory instead """

        # icon path for app bundle (tmp) or non-bundled package (cwd)
        cwd = os.path.join(os.path.dirname(os.path.realpath(__file__)), self.REL_PATH)
        tmp = os.path.join(sys._MEIPASS, self.REL_PATH) if hasattr(sys, '_MEIPASS') else None
        fp = cwd if tmp is None else tmp

        if sys.platform == 'linux':
            # load icon on linux
            logo = tk.PhotoImage(master=self, file=fp)
            self.wm_iconphoto(True, logo)

        elif sys.platform == 'win32':
            # generate blank window icon
            _, ICON_PATH = mkstemp()
            with open(ICON_PATH, 'wb') as icon_file:
                icon_file.write(ICON)

            # load icon on Windows
            fp = fp.replace('gif', 'ico')
            fp = fp if os.path.exists(fp) else ICON_PATH
            self.iconbitmap(fp)
            self.wm_iconbitmap(default=fp)


def main_app():

    # instantiate object
    app_win = PlenopticamApp()
    # make not resizable
    app_win.resizable(width=0, height=0)
    # run
    app_win.mainloop()


if __name__ == "__main__":

    main_app()



import numpy as np


class LineFitter(object):

    def __init__(self, *args, **kwargs):
        super(LineFitter, self).__init__()

        # use config and status if passed
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else None
        self.sta = kwargs['sta'] if 'sta' in kwargs else None

        self._grid_fit = list()

        # take coordinates as input argument
        if 'coords_list' in kwargs:
            self._coords_list = np.asarray(kwargs['coords_list'])
        elif self.cfg.mic_list in self.cfg.calibs:
            self._coords_list = np.asarray(self.cfg.calibs[self.cfg.mic_list])
        else:
            raise BaseException('Coordinate list not provided.')

        if hasattr(self, '_coords_list'):
            self._MAX_Y = int(max(self._coords_list[:, 2])+1)   # add one to compensate for index zero
            self._MAX_X = int(max(self._coords_list[:, 3])+1)   # add one to compensate for index zero
            self._pat_type = self.cfg.calibs[self.cfg.pat_type] if self.cfg.pat_type in self.cfg.calibs else 'rec'

    def main(self, deg=1):

        # check for minimum grid resolution
        if self._pat_type == 'rec' and self._MAX_Y > 3 and self._MAX_X > 3 or self._MAX_Y > 5 and self._MAX_X > 5:
            self.comp_grid_fit(deg=deg)
        else:
            print('Skip grid fitting as grid point number is too little')
            self._grid_fit = np.array(self._coords_list)

        return True

    def comp_grid_fit(self, deg=1):
        """ perform fitting for two dimensional grid coordinates """

        # print status
        self.sta.status_msg('Grid fitting', self.cfg.params[self.cfg.opt_prnt])

        odd = 0     # hexagonal shift direction

        # iterate through rows
        for ly in range(self._MAX_Y):

            # check interrupt status
            if self.sta.interrupt:
                return False

            # fit line of all coordinates in current row
            coords_row = self._coords_list[self._coords_list[:, 2] == ly][:, :2]
            coeffs_hor = self.line_fitter(coords_row, axis=1, deg=deg)

            # iterate through column
            for lx in range(self._MAX_X):

                # select coordinates
                coords_col = self._coords_list[self._coords_list[:, 3] == lx][:, :2]

                # consider hexagonal pattern
                if self._pat_type == 'hex':
                    # fit line of current column coordinates (omit every other coordinate pair)
                    coeffs_ver = self.line_fitter(coords_col[odd::2], axis=1, deg=deg)

                else:
                    # fit line of current column coordinates
                    coeffs_ver = self.line_fitter(coords_col, axis=1, deg=deg)

                # compute intersection of row and column line
                new_coords = list(self.line_intersect(coeffs_hor, coeffs_ver))

                # original reference coordinate
                ref_coords = self._coords_list[(self._coords_list[:, 2] == ly) & (self._coords_list[:, 3] == lx)]

                # put result to new coordinate list (if reference exists)
                self._grid_fit.append(new_coords + [ly, lx]) if ref_coords.size != 0 else None

            # switch hexagonal shift direction
            odd = not odd if self._pat_type == 'hex' else odd

            # print progress status
            percentage = (ly + 1) / self._MAX_Y * 100
            self.sta.progress(percentage, opt=self.cfg.params[self.cfg.opt_prnt]) if self.sta else None

        self._grid_fit = np.array(self._grid_fit)

        return self._grid_fit

    @staticmethod
    def line_intersect(coeffs_hor, coeffs_ver):
        """ compute line intersection based on Vandermonde algebra """

        deg = len(coeffs_hor)-1 if len(coeffs_hor) == len(coeffs_ver) else 0

        if deg < 3:
            # intersection of linear functions
            matrix = np.array([[1, *-coeffs_hor[1:]], [1, *-coeffs_ver[1:]]])
            vector = np.array([coeffs_hor[0], coeffs_ver[0]])
        else:
            # intersection of non-linear functions
            raise BaseException('Function with degree %s is not supported' % deg)

        return np.dot(np.linalg.pinv(matrix), vector)

    def line_fitter(self, coords, axis=0, deg=1):
        """ estimate equation fit of 2-D coordinates belonging to the same row or column via least squares method """

        # feed into system of equations
        A = self.compose_vandermonde_1d(coords[:, axis], deg=deg)
        b = coords[:, 1-axis]

        # solve for a least squares estimate via pseudo inverse and coefficients in b
        coeffs = np.dot(np.linalg.pinv(A), b)

        return coeffs

    @staticmethod
    def compose_vandermonde_1d(x, deg=1):
        if deg == 1:
            return np.array([np.ones(len(x)), x]).T
        elif deg == 2:
            return np.array([np.ones(len(x)), x, x ** 2]).T
        elif deg == 3:
            return np.array([np.ones(len(x)), x, x ** 2, x ** 3]).T

    @property
    def grid_fit(self):

        # convert array indices to integer
        self._grid_fit[:, 2] = self._grid_fit[:, 2].astype('int')
        self._grid_fit[:, 3] = self._grid_fit[:, 3].astype('int')

        return self._grid_fit.tolist()



from __future__ import print_function

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.misc.type_checks import isfloat
import threading


class PlenopticamStatus(object):

    def __init__(self):

        # private instance variables
        self._prog_var = 0
        self._stat_var = ''
        self._interrupt = threading.Event()
        self._error = False

        # observer lists to bind to
        self._prog_observers = []
        self._stat_observers = []
        self._interrupt_observers = []

        # public instance variables
        self.prog_var = 0
        self.prog_opt = True
        self.stat_var = ''
        self.interrupt = False
        self.error = False

    def progress(self, x, opt=False):

        if isfloat(x) and not self.interrupt:
            # only update if there is a significant (1.0%) change to prevent time-consuming re-prints
            curr_prog_var = round(x, 0)
            if curr_prog_var != self.prog_var:
                self.prog_var = int(round(x, 0))
                if x == 100:
                    self.prog_var = 'Finished'

            # console print (if option set)
            if opt and self.prog_opt:
                print('\r Progress: {:2.0f}%'.format(float(x)), end='')

                if x == 100:
                    print('\r Progress: Finished \n')
        elif self.interrupt:
            self.prog_var = ''
        elif x is None:
            self.prog_var = 'Processing'
        else:
            self.prog_var = str(x)

        return True

    def status_msg(self, msg='', opt=False):

        if not self.interrupt:
            self.stat_var = msg

            # reset progress bar
            self.progress(None, opt=False)

            # console print (if option set)
            if opt:
                print('\n', msg)
        else:
            self.stat_var = 'Canceling' if self._error is False else self.stat_var
            self.progress('', opt=False)

            return True

    def validate(self, checklist=None, msg=None):

        checklist = checklist if checklist is not None else list()

        for el in checklist:
            if el:
                return True

        # list elements are all "None" or empty
        self.status_msg(msg, msg is not None)
        self.error = True

        return False

    # event trigger on parameter change (according to observer pattern)
    @property
    def stat_var(self):
        return self._stat_var

    @stat_var.setter
    def stat_var(self, msg):
        self._stat_var = msg
        for callback in self._stat_observers:
            callback(self._stat_var)

    @property
    def prog_var(self):
        return self._prog_var

    @prog_var.setter
    def prog_var(self, val):
        self._prog_var = val
        for callback in self._prog_observers:
            callback(self._prog_var)

    def bind_to_prog(self, callback):
        self._prog_observers.append(callback)

    def bind_to_stat(self, callback):
        self._stat_observers.append(callback)

    # interrupt getter
    @property
    def interrupt(self):
        return self._interrupt.is_set()

    # interrupt change detection
    @interrupt.setter
    def interrupt(self, val):
        if val:

            # set interrupt
            self._interrupt.set()

            for callback in self._interrupt_observers:
                callback()

            # display stop message and reset progress bar
            self.status_msg()

        elif not val:
            # reset interrupt status
            self._interrupt.clear()

    def bind_to_interrupt(self, callback):
        self._interrupt_observers.append(callback)

    # error getter
    @property
    def error(self):
        return self._error

    # error setter
    @error.setter
    def error(self, val):
        if val:
            self._error = True
            self.interrupt = True
        else:
            self._error = False



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from plenopticam import misc
from plenopticam.lfp_extractor import LfpViewpoints
from color_space_converter import hsv_conv, yuv_conv, rgb2gry


class LfpContrast(LfpViewpoints):

    def __init__(self, p_lo=None, p_hi=None, *args, **kwargs):
        super(LfpContrast, self).__init__(*args, **kwargs)

        self.p_lo = p_lo if p_lo is not None else 0.0
        self.p_hi = p_hi if p_hi is not None else 1.0

        self.ref_img = kwargs['ref_img'] if 'ref_img' in kwargs else self.central_view

        # internal variables
        self._contrast, self._brightness = (1., 1.)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # auto contrast balance
        if self.cfg.params[self.cfg.opt_cont] and not self.sta.interrupt:
            self.p_hi, self.p_lo = (1, 0)
            self.con_bal()
        # auto white balance
        if self.cfg.params[self.cfg.opt_awb_] and not self.sta.interrupt:
            self.p_hi, self.p_lo = (1, 0)
            self.wht_bal()
        # auto saturation
        if self.cfg.params[self.cfg.opt_sat_] and not self.sta.interrupt:
            self.p_hi, self.p_lo = (1, 0)
            self.sat_bal()

        # automatic histogram alignment
        self._vp_img_arr = self.auto_hist_align(img=self._vp_img_arr, ref_img=self.central_view, opt=True)

        # gamma correction
        self._vp_img_arr = misc.GammaConverter().srgb_conv(img=self._vp_img_arr)

        return True

    def con_bal(self):

        # find extrema from reference image
        self.ref_img = yuv_conv(self.central_view)[..., 0]
        max = self.ref_img.max()
        min = self.ref_img.min()

        # convert to yuv space
        self.proc_vp_arr(yuv_conv, msg='Convert to YUV')

        # boost luminance channel
        self.sta.status_msg(msg='Contrast balance', opt=self.cfg.params[self.cfg.opt_prnt])
        self._vp_img_arr[..., 0] = misc.Normalizer(self._vp_img_arr[..., 0]).type_norm(max=max, min=min)
        self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt])

        # convert to rgb space
        self.proc_vp_arr(yuv_conv, inverse=True, msg='Convert to RGB')

        return True

    def sat_bal(self):

        # convert to hsv space
        self.proc_vp_arr(hsv_conv, msg='Convert to HSV')

        # boost saturation channel
        self.sta.status_msg(msg='Saturation balance', opt=self.cfg.params[self.cfg.opt_prnt])
        self.vp_img_arr[..., 1] *= 1.1
        self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt])

        # convert to rgb space
        self.proc_vp_arr(hsv_conv, inverse=True, msg='Convert to RGB')

        return True

    def wht_bal(self, method=None, msg_opt=True):

        # status update
        if msg_opt:
            msg = 'Auto white balance' if self.p_hi != 1 else 'Color adjustment'
            self.sta.status_msg(msg=msg, opt=self.cfg.params[self.cfg.opt_prnt])
            self.sta.progress(0, opt=self.cfg.params[self.cfg.opt_prnt])

        ch_num = self.vp_img_arr.shape[-1] if len(self.vp_img_arr.shape) > 4 else 3
        for i in range(ch_num):
            if method is None:

                # channel selection
                ref_ch = self.ref_img[..., i]
                img_ch = self.vp_img_arr[..., i]

                # define level limits
                min = np.percentile(ref_ch, self.p_lo*100)
                max = np.percentile(ref_ch, self.p_hi*100)

                # normalization of color channel
                self.vp_img_arr[..., i] = misc.Normalizer(img_ch, min=min, max=max).uint16_norm()

            else:
                # brightness and contrast method
                self.set_stretch(ref_ch=self.ref_img[..., i])
                self.apply_stretch(ch=i)

            # status update
            if msg_opt:
                self.sta.progress((i+1)/ch_num*100, opt=self.cfg.params[self.cfg.opt_prnt])

        return True

    def channel_bal(self):

        # status update
        self.sta.status_msg(msg='Contrast balance', opt=self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, opt=self.cfg.params[self.cfg.opt_prnt])

        ch_num = self.vp_img_arr.shape[-1] if len(self.vp_img_arr.shape) > 4 else 3

        min = float('Inf')
        max = 0.
        for i in range(ch_num):
            min = np.min([min, np.percentile(self.ref_img[..., i], self.p_lo * 100)])
            max = np.max([max, np.percentile(self.ref_img[..., i], self.p_hi * 100)])

        # normalization of color channel
        self.vp_img_arr = misc.Normalizer(self.vp_img_arr, min=min, max=max).uint16_norm()

        # status update
        self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt])

        return True

    def stretch_contrast(self):

        # estimate contrast and brightness via least-squares method
        self.set_stretch(ref_ch=yuv_conv(self.central_view)[..., 0])

        self.proc_vp_arr(yuv_conv, msg='Convert to YUV')
        self.proc_vp_arr(self.apply_stretch, ch=0, msg='Contrast balance')
        self.proc_vp_arr(yuv_conv, inverse=True, msg='Convert to RGB')

        # status update
        self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt])

        return True

    def set_stretch(self, ref_ch, val_lim=None):
        """ according to https://stackoverflow.com/questions/9744255/instagram-lux-effect/9761841#9761841 """

        # estimate contrast und brightness parameters (by default: first channel only)
        val_lim = 2**16-1 if not val_lim else val_lim
        h = np.histogram(ref_ch, bins=val_lim)[0]
        H = np.cumsum(h)/float(np.sum(h))
        try:
            px_lo = self.find_x_given_y(self.p_lo, np.arange(val_lim), H)
            px_hi = self.find_x_given_y(self.p_hi, np.arange(val_lim), H)
        except:
            px_lo = 0
            px_hi = val_lim
        A = np.array([[px_lo, 1], [px_hi, 1]])
        b = np.array([0, val_lim])
        self._contrast, self._brightness = np.dot(np.linalg.inv(A), b)

        return self._contrast, self._brightness

    @staticmethod
    def find_x_given_y(value, x, y, tolerance=1e-4):

        i = 0
        iter_max = 10
        arr = np.array([])
        while arr.size == 0 and i != iter_max:
            arr = np.array([(xi, yi) for (xi, yi) in zip(x, y) if abs(yi - value) <= tolerance])
            i += 1
            tolerance = 10**(-4+i)

        found_y = np.mean(arr.T[0]) if i != iter_max else round(value)

        return found_y

    def apply_stretch(self, img=None, ch=None):
        """ contrast and brightness rectification for provided RGB image """

        img = img if img is not None else self.vp_img_arr
        ch = ch if ch is not None else 0

        # convert to float
        f = img.astype(np.float64)

        # perform auto contrast (by default: "lum" channel only)
        img[..., ch] = self._contrast * f[..., ch] + self._brightness

        # clip to input extrema to remove contrast outliers
        #img[img < f.min()] = f.min()
        #img[img > f.max()] = f.max()

        return img

    def set_stretch_hsv(self):

        # use luminance channel for parameter analysis
        ref_img = hsv_conv(self.ref_img)
        self.set_stretch(ref_ch=ref_img[..., 1]*(2**16-1))

    def apply_stretch_hsv(self, img):
        ''' saturation '''

        # color model conversion
        hsv = hsv_conv(img)

        # apply histogram stretching to saturation channel only
        hsv[..., 1] *= (2**16-1)
        self.apply_stretch(hsv, ch=1)
        hsv[..., 1] /= (2**16-1)

        # color model conversion
        rgb = hsv_conv(hsv, inverse=True)

        return rgb

    @staticmethod
    def auto_hist_align(img, ref_img, opt=None):

        if opt:
            p_lo, p_hi = (0.005, 99.9)
            min_perc = np.percentile(rgb2gry(ref_img), p_lo)
            max_perc = np.percentile(ref_img, p_hi)
        else:
            p_lo, p_hi = (0.5, 99.9)
            min_perc = np.percentile(ref_img, p_lo)
            max_perc = np.percentile(ref_img, p_hi)

        img = misc.Normalizer(img, min=min_perc, max=max_perc).type_norm()

        return img

    def thresh_hist_stretch(self, th=2e-10, bins=2**16-1):

        h = np.histogram(self.central_view, bins=bins)
        hn = h[0] / h[0].sum()
        x_vals = np.where(hn / len(hn) > th)[0] / bins

        hs = np.diff(h[0][::128] / h[0][::128].sum())
        s_vals = np.where(hs > 1.5e-4)[0] / (bins / 128)

        #img = misc.Normalizer(self.central_view.copy(), min=x_vals[1], max=self.central_view.max()).type_norm()
        self.proc_vp_arr(misc.Normalizer().type_norm, msg='Histogram crop', min=x_vals[1], max=1)

        return True

    def post_lum(self, ch=None):

        self.vp_img_arr = misc.Normalizer(self.vp_img_arr).uint16_norm()

        # channel selection
        ch = ch if ch is not None else 0
        ref_ch = yuv_conv(self.central_view)[..., ch]

        # define level limits
        self._min = np.percentile(ref_ch, self.p_lo*100)
        self._max = np.percentile(ref_ch, self.p_hi*100)

        self.proc_vp_arr(self.lum_norm, msg='Luminance normalization')

    def lum_norm(self, img, ch=None):

        # set default channel
        ch = ch if ch is not None else 0

        # RGB to YUV conversion
        img = yuv_conv(img)

        # normalization of Y (luminance channel)
        img = misc.Normalizer(img[..., ch], min=self._min, max=self._max).uint16_norm()

        # YUV to RGB conversion
        img = yuv_conv(img, inverse=True)

        return img



import numpy as np
from scipy.optimize import leastsq, least_squares
from typing import Union

l2_norm_eucl = lambda c_meas, c_grid: np.sqrt(np.sum((c_meas - c_grid) ** 2, axis=1))
l1_norm_eucl = lambda c_meas, c_grid: np.sum(np.abs(c_meas - c_grid), axis=1)
l2_norm_elem = lambda c_meas, c_grid: np.square(c_meas - c_grid)
l1_norm_elem = lambda c_meas, c_grid: np.abs(c_meas - c_grid)


class GridFitter(object):

    def __init__(self, *args, **kwargs):
        super(GridFitter, self).__init__()

        # use config and status if passed
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else None
        self.sta = kwargs['sta'] if 'sta' in kwargs else None

        # internal variables
        self._grid_fit = np.array([])
        self._coeffs = np.identity(3).flatten()
        self._pat_type = kwargs['pat_type'] if 'pat_type' in kwargs else 'rec'
        self._ptc_mean = kwargs['ptc_mean'] if 'ptc_mean' in kwargs else np.array([1, 1])
        self._hex_odd = kwargs['hex_odd'] if 'hex_odd' in kwargs else 0
        self._arr_shape = np.array(kwargs['arr_shape'])[:2] if 'arr_shape' in kwargs else np.array([0, 0])
        self._affine = kwargs['affine'] if 'affine' in kwargs else False
        self._compose = kwargs['compose'] if 'compose' in kwargs else False
        self._flip_yx = kwargs['flip_xy'] if 'flip_xy' in kwargs else False
        self._normalize = kwargs['normalize'] if 'normalize' in kwargs else False
        self.z_dist = kwargs['z_dist'] if 'z_dist' in kwargs else 1.0

        # regression settings
        self.penalty_enable = kwargs['penalty_enable'] if 'penalty_enable' in kwargs else False

        # take coordinate array and its properties as input
        if 'cfg' in kwargs and self.cfg.mic_list in self.cfg.calibs:
            self._coords_list = np.asarray(self.cfg.calibs[self.cfg.mic_list])
            self._pat_type = self.cfg.calibs[self.cfg.pat_type] if self.cfg.pat_type in self.cfg.calibs else 'rec'
            self._ptc_mean = self.cfg.calibs[self.cfg.ptc_mean] if self.cfg.ptc_mean in self.cfg.calibs else [1, 1]
            self._hex_odd = self.estimate_hex_odd(self._coords_list) if self._pat_type == 'hex' else self._hex_odd
        if 'coords_list' in kwargs:
            self._coords_list = np.asarray(kwargs['coords_list'])
            if self._coords_list.shape[1] == 4:
                self._ptc_mean = np.mean(np.abs(np.diff(self._coords_list[self._coords_list[:, 2] == 0][:, 1])))
        elif len(args) > 0:
            self._coords_list = np.asarray(args[0])

        if hasattr(self, '_coords_list'):
            # add one to list index to compensate for index zero
            self._MAX_Y = int(max(self._coords_list[:, 2])+1)
            self._MAX_X = int(max(self._coords_list[:, 3])+1)
        else:
            self._MAX_Y, self._MAX_X = 2, 2

        self._ptc_mean = [self._ptc_mean]*2 if isinstance(self._ptc_mean, (int, float, np.float64)) else self._ptc_mean

    def main(self):

        # check for minimum grid resolution
        if self._MAX_Y > 3 and self._MAX_X > 3 or self._MAX_Y > 5 and self._MAX_X > 5:
            self.comp_grid_fit()
        else:
            print('Skip grid fitting as grid point number is too little')
            self._grid_fit = np.array(self._coords_list)

        if hasattr(self, 'cfg') and hasattr(self.cfg, 'calibs'):
            self.cfg.calibs[self.cfg.mic_list] = self._grid_fit

        return True

    def coeff_fit(self, coords_list=None, euclid_opt=True):
        """ perform two dimensional grid regression """

        self._coords_list = coords_list if coords_list is not None else self._coords_list

        # initial fit parameters
        cy_s, cx_s = self._arr_shape/2 if hasattr(self, '_arr_shape') else [0, 0]
        sy_s, sx_s = self._ptc_mean if hasattr(self, '_ptc_mean') else [1, 1]
        p_init = np.diag([sy_s, sx_s, 1])
        p_init[:2, -1] = np.array([cy_s, cx_s])
        p_init = p_init.flatten()[:8]
        beta = 1 if self.penalty_enable else 0

        # LMA fit: executes least-squares regression for optimization of initial parameters
        try:
            self._coeffs = leastsq(self.cost_fun, p_init, args=(self._coords_list, beta, euclid_opt))[0]
        except:
            # newer interface for LMA
            self._coeffs = least_squares(self.cost_fun, p_init.flatten(),
                                            jac='2-point', args=(self._coords_list, beta, euclid_opt), method='lm').x

    def comp_grid_fit(self):
        """ perform two dimensional grid regression and return fitted grid """

        # print status
        self.sta.status_msg('Grid fitting', self.cfg.params[self.cfg.opt_prnt]) if self.sta else None

        # perform the grid regression
        self.coeff_fit()

        # generate initial grid
        grid_init = self.grid_gen(dims=[self._MAX_Y, self._MAX_X],
                                  pat_type=self._pat_type,
                                  hex_odd=self._hex_odd,
                                  normalize=self._normalize
                                  )

        # project grid from final estimates
        self._grid_fit = self.apply_transform(self._coeffs, grid_init, self._affine, self._flip_yx, self.z_dist)

        # print status
        self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt]) if self.sta else None

        return np.array(self._grid_fit)

    def cost_fun(self, p, centroids, beta=0, euclid_opt=True):

        # generate grid points
        grid_pts = self.grid_gen(dims = [self._MAX_Y, self._MAX_X],
                                 pat_type = self._pat_type,
                                 hex_odd = self._hex_odd,
                                 normalize = self._normalize
                                 )

        # generate projection parameters
        p = self.compose_p(p) if self._compose else p

        # transform grid points
        grid = self.apply_transform(p, grid_pts, self._affine, self._flip_yx, self.z_dist)

        # choose euclidian or element-wise norm (the latter yields twice as many points)
        norm_fun = l2_norm_eucl if euclid_opt else l2_norm_elem

        # mask non-existing points
        idxs = np.ones(grid.shape[0], dtype='bool')
        if centroids.shape[0] != grid.shape[0]:
            # 4 corner fitting
            if centroids.shape[0] == 4:
                idxs = np.where((grid[:, 2] == 0) & (grid[:, 3] == 0) |
                                (grid[:, 2] == 0) & (grid[:, 3] == self._MAX_X-1) |
                                (grid[:, 2] == self._MAX_Y-1) & (grid[:, 3] == 0) |
                                (grid[:, 2] == self._MAX_Y-1) & (grid[:, 3] == self._MAX_X-1)
                                )

        # compute loss
        try:
            loss = norm_fun(centroids[:, :2], grid[:, :2][idxs])
            loss += beta * self._regularizer(centroids.copy(), grid) if beta > 0 else 0
        except ValueError:
            err_msg = 'Grid index mismatch'
            if self.sta is None:
                raise Exception(err_msg)
            else:
                self.sta.status_msg(err_msg, True)
                self.sta.error = True
                return False

        return loss.flatten()
    
    @staticmethod
    def apply_transform(p, grid: np.ndarray, affine: bool = False, flip_xy: bool = False, z_dist: float = 1.):
        """ transformation """

        # append 1 to 8 parameters vector
        p = np.array([*p, 1]) if len(p) == 8 else np.array(p)

        # reshape vector to 3x3 matrix
        pmat = p.reshape(3, 3) if len(p) == 9 else p

        # transformation constraint
        pmat[-1, :] = np.array([0, 0, 1]) if affine else np.array([*pmat[-1, :2], 1])

        # form points matrix adding vector of ones
        pts = np.concatenate((grid[:, :2].T, z_dist*np.ones(len(grid))[np.newaxis, :]), axis=0)

        # flip x and y coordinates
        pts[:2, :] = pts[:2, :][::-1] if flip_xy else pts[:2, :]

        # transform points
        prj = np.dot(pmat, pts)

        # flip y and x coordinates
        prj[:2, :] = prj[:2, :][::-1] if flip_xy else prj[:2, :]

        # project z values (if available)
        grid[:, :2] = np.divide(prj[:2, :], prj[2, :], out=np.zeros(prj[:2, :].shape), where=prj[2, :] != 0).T[:, :2]

        return grid

    @staticmethod
    def _regularizer(c_meas, c_grid, div=22):

        assert c_meas.shape[-1] == 4, 'Regularizer requires 4 columns in the 2-D array'

        dim = int(np.max(c_meas[:, 3]))
        pitch = np.mean(np.diff(c_meas[:dim, 1]))

        c_meas[:, 0] -= np.mean(c_meas[:, 0])
        c_meas[:, 1] -= np.mean(c_meas[:, 1])
        c_grid[:, 0] -= np.mean(c_grid[:, 0])
        c_grid[:, 1] -= np.mean(c_grid[:, 1])

        diff = np.abs(c_meas) - np.abs(c_grid)

        # penalty boundary is further pushed by additional fraction
        diff += pitch / div

        diff[:, 0][diff[:, 0] < 0] = 0
        diff[:, 1][diff[:, 1] < 0] = 0
        diff = np.sum(diff, axis=1)

        return diff

    @staticmethod
    def grid_gen(dims: Union[int, int] = None, pat_type: str = None, hex_odd: bool = None, normalize: int = 0):
        """ generate grid of points """

        # set default values
        dims = [5, 5] if dims is None else dims
        pat_type = 'rec' if pat_type is None else pat_type
        hex_odd = 0 if hex_odd is None else hex_odd

        assert pat_type == 'rec' or pat_type == 'hex' or pat_type == 'pseudo-hex', 'Grid pattern type not recognized.'

        # create point coordinates
        y_range = np.linspace(0, dims[0]-1, dims[0])
        x_range = np.linspace(0, dims[1]-1, dims[1])
        y_grid, x_grid = np.meshgrid(y_range, x_range)

        # create point index labels
        y_idcs, x_idcs = np.meshgrid(np.linspace(0, dims[0]-1, dims[0]),
                                     np.linspace(0, dims[1]-1, dims[1]))

        # shift rows horizontally if grid is hexagonal
        if pat_type == 'hex':
            # vertical shrinkage
            y_grid *= np.sqrt(3) / 2
            # horizontal hex shifting
            x_grid[:, int(hex_odd)::2] += .5
        elif pat_type == 'pseudo-hex':
            # horizontal hex shifting
            x_grid[:, int(hex_odd)::2] += .5

        # normalize grid
        if normalize > 0:
            norm_div = (max(x_grid.max(), y_grid.max()),)*2 if normalize == 1 else (x_grid.max(), y_grid.max())
            y_grid, x_grid = y_grid / norm_div[0], x_grid / norm_div[1]

        # put grid to origin
        y_grid -= y_grid.max()/2
        x_grid -= x_grid.max()/2

        pts_arr = np.vstack(np.dstack([y_grid.T, x_grid.T, y_idcs.T, x_idcs.T]))

        return pts_arr

    @staticmethod
    def compose_p(p):

        p = p.flatten()
        rmat = GridFitter.euler2mat(theta_x=p[7], theta_y=p[6], theta_z=p[3])
        kmat = np.diag([p[0], p[4], 1])
        kmat[:2, -1] = np.array([p[2], p[5]])
        p = np.dot(rmat, kmat).flatten()

        return p

    @staticmethod
    def estimate_hex_odd(coords_list):

        # find smallest lens indices (more robust than finding lens index zero)
        min_idx_y, min_idx_x = min(coords_list[:, 2]), min(coords_list[:, 3])

        # take the most upper two points in hexagonal grid
        pt_00 = coords_list[(coords_list[:, 3] == min_idx_x) & (coords_list[:, 2] == min_idx_y)][0]
        pt_10 = coords_list[(coords_list[:, 3] == min_idx_x) & (coords_list[:, 2] == min_idx_y + 1)][0]

        # compare x-coordinates to determine the direction of hexagonal shift alternation
        hex_odd = 1 if pt_00[1] < pt_10[1] else 0

        return hex_odd

    @staticmethod
    def euler2mat(theta_x: float = 0, theta_y: float = 0, theta_z: float = 0):
        """
        Creation of a rotation matrix from three angles in radians
        """

        # matrix for counter-clockwise rotation around x-y-z axes
        rmat_z = np.array([[np.cos(theta_z), -np.sin(theta_z), 0],
                           [np.sin(theta_z), np.cos(theta_z), 0],
                           [0, 0, 1]]
                          )

        rmat_y = np.array([[np.cos(theta_y), 0, np.sin(theta_y)],
                           [0, 1, 0],
                           [-np.sin(theta_y), 0, np.cos(theta_y)]]
                          )

        rmat_x = np.array([[1, 0, 0],
                           [0, np.cos(theta_x), -np.sin(theta_x)],
                           [0, np.sin(theta_x), np.cos(theta_x)]]
                          )

        rmat_3 = np.dot(np.dot(rmat_x, rmat_y), rmat_z)

        return rmat_3

    @staticmethod
    def decompose(pmat, scale=True):
        """
        https://www.robots.ox.ac.uk/~vgg/hzbook/code/vgg_multiview/vgg_KR_from_P.m
        """

        # reshape potential vector to 3x3 matrix
        pmat = np.array(pmat).reshape(3, 3) if len(pmat.shape) == 1 and np.array(pmat).size == 9 else pmat

        n = pmat.shape[0] if len(pmat.shape) == 2 else np.sqrt(pmat.size)
        rmat, kmat = np.linalg.qr(pmat.reshape(n, -1), mode='reduced')

        if scale:
            kmat = kmat / kmat[n-1, n-1]
            if kmat[0, 0] < 0:
                D = np.diag([-1, -1, *np.ones(n-2)])
                kmat = np.dot(D, kmat)
                rmat = np.dot(rmat, D)

        tvec = -1*np.linalg.lstsq(pmat[:, :n], pmat[:, -1]) if pmat.shape[1] == 4 else np.zeros(n)

        return rmat, kmat, tvec

    @staticmethod
    def mat2euler(rmat):
        """
        https://www.geometrictools.com/Documentation/EulerAngles.pdf
        """

        if rmat[0, 2] < 1:
            if rmat[0, 2] > -1:
                theta_y = np.arcsin(rmat[0, 2])
                theta_x = np.arctan2(-rmat[1, 2], rmat[2, 2])
                theta_z = np.arctan2(-rmat[0, 1], rmat[0, 0])
            else:
                theta_y = -np.pi / 2
                theta_x = -np.arctan2(rmat[1, 0], rmat[1, 1])
                theta_z = 0
        else:
            theta_y = np.pi / 2
            theta_x = np.arctan2(rmat[1, 0], rmat[1, 1])
            theta_z = 0

        return np.array([theta_x, theta_y, theta_z])

    @property
    def grid_fit(self):

        # convert array indices to integer
        self._grid_fit[:, 2] = self._grid_fit[:, 2].astype('int')
        self._grid_fit[:, 3] = self._grid_fit[:, 3].astype('int')

        return self._grid_fit.tolist()

    @property
    def pmat(self):
        self._coeffs = np.array([*self._coeffs, 1]) if len(self._coeffs) == 8 else self._coeffs
        if self._compose:
            return self.compose_p(self._coeffs).reshape(3, 3)
        else:
            return self._coeffs.reshape(3, 3)

    @pmat.setter
    def pmat(self, pmat: np.ndarray):
        if pmat.size == 9:
            self._coeffs = pmat.flatten()
        else:
            raise Exception('Only a 9-vector or 3x3 matrix is accepted')

    @property
    def rmat(self):
        return self.decompose(self.pmat)[0]

    @property
    def kmat(self):
        return self.decompose(self.pmat)[1]

    @property
    def tvec(self):
        return self.decompose(self.pmat)[2]



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import PlenopticamStatus

import numpy as np


class ClsRefoSlices(object):

    def __init__(self, *args, **kwargs):

        self._rf_img_arr = kwargs['rf_img_arr'].astype('float64') if 'rf_img_arr' in kwargs else None
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else PlenopticamConfig()
        self.sta = kwargs['sta'] if 'sta' in kwargs else PlenopticamStatus()
        self._M = self.cfg.params[self.cfg.ptc_leng]
        self._C = self._M // 2

    @property
    def rf_img_arr(self):
        return self._rf_img_arr

    @rf_img_arr.setter
    def rf_img_arr(self, rf_img_arr):
        self._rf_img_arr = rf_img_arr

    def proc_refo_arr(self, fun, **kwargs):
        ''' process viewpoint images based on provided function handle and argument data '''

        # percentage indices for tasks having sub-processes
        iter_num = kwargs['iter_num'] if 'iter_num' in kwargs else 0
        iter_tot = kwargs['iter_tot'] if 'iter_tot' in kwargs else 1

        # status message handling
        if iter_num == 0:
            msg = kwargs['msg'] if 'msg' in kwargs else 'Refocusing process'
            self.sta.status_msg(msg, self.cfg.params[self.cfg.opt_prnt])

        args = self.remove_proc_keys(kwargs, data_type=list)

        try:
            if len(self.rf_img_arr.shape) == 4:
                for j in range(self._rf_img_arr.shape[0]):
                    self._rf_img_arr[j, ...] = fun(self._rf_img_arr[j, ...], *args)

                    # progress update
                    percent = (j*self._rf_img_arr.shape[1]+1)/np.dot(*self._rf_img_arr.shape[:2])
                    percent = percent / iter_tot + iter_num / iter_tot
                    self.sta.progress(percent*100, self.cfg.params[self.cfg.opt_prnt])
        except:
            if len(self.rf_img_arr.shape) != 4:
                raise NotImplementedError



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import os
from scipy.interpolate import interp1d

from plenopticam import misc
from plenopticam.lfp_extractor import LfpViewpoints
from plenopticam.lfp_aligner.lfp_local_resampler import LfpLocalResampler


class HexCorrector(LfpViewpoints):

    def __init__(self, *args, **kwargs):
        super(HexCorrector, self).__init__(*args, **kwargs)

        self.ref_img = kwargs['ref_img'] if 'ref_img' in kwargs else self.central_view
        self.method = kwargs['method'] if 'method' in kwargs else 'cubic'

        if self.cfg.calibs[self.cfg.mic_list] is not None:
            # analyse
            self.hex_odd = LfpLocalResampler.get_hex_direction(np.asarray(self.cfg.calibs[self.cfg.mic_list]))
        else:
            # reset pattern type to skip hex correction process
            self.cfg.calibs[self.cfg.pat_type] = 'rec'

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # remove hexagonal artifact
        if self.cfg.calibs[self.cfg.pat_type] == 'hex':
            self.proc_vp_arr(self.ver_hex_bulge, msg='Hexagonal artifact removal')

        # normalize light-field
        self.vp_img_arr = misc.Normalizer(self.vp_img_arr).uint16_norm()

        return True

    def hex_interp_1d(self, img):

        hex_stretch = 2 / np.sqrt(3)
        m, n, ch = img.shape if len(img.shape) == 3 else img.shape + (1,)
        s_check = [0, 1] if self.hex_odd else [1, 0]
        final = np.zeros([m, int(np.round(n * hex_stretch)), ch])
        new_l = int(np.round(img.shape[1] * hex_stretch))

        img = img.astype('float')/img.max()
        for p in range(ch):
            for j in range(2):
                s = -0.5 * s_check[j]
                for i in range(img.shape[0])[j::2]:
                    fun = interp1d(np.arange(len(img[i, :, p])), img[i, :, p], kind=self.method, fill_value="extrapolate")
                    final[i, ..., p] = fun(np.linspace(0, len(img[i, :, p])-1, new_l)+s)

        return final

    def ver_hex_bulge(self, img):

        # variable init
        img = np.divide(img.astype('float'), img.max(), out=np.zeros_like(img), where=img != 0)
        m, n, ch = img.shape if len(img.shape) == 3 else img.shape + (1,)

        # j for shifted and i represents truth (unshifted)
        i, j = [1, 0] if self.hex_odd else [0, 1]

        # gradient between shifted and unshifted (responsive to hex artifacts)
        if img[i::2, ...].shape[0] == img[j::2, ...].shape[0]:
            dif = img[j::2, ...]-img[i::2, ...]
        elif img[j::2, ...].shape[0] == img[i::2, ...].shape[0]+1:
            dif = img[j::2, ...][:-1]-img[i::2, ...]
        elif img[j::2, ...].shape[0]+1 == img[i::2, ...].shape[0]:
            dif = img[j::2, ...] - img[i::2, ...][:-1]
        else:
            self.sta.error('Image dimension mismatch')
            return False

        # gradient between unshifted (responsive to real object edges)
        edg = img[i::2, ...][:-1]-img[i+2::2, ...]
        edg *= 0.75

        # absolute values
        dif = abs(dif)
        edg = abs(edg)

        # (don't throw away channel information)
        dif = dif.sum(2)
        edg = edg.sum(2)

        # deduct actual edges
        karr = dif - edg if dif.shape[0] == edg.shape[0] else dif[:-1, ...] - edg
        karr = np.divide(karr, karr.max(), out=np.zeros_like(karr), where=karr != 0)
        karr[karr < 0] = 0
        if self.cfg.params[self.cfg.opt_dbug]:
            misc.save_img_file(karr, os.path.join(self.cfg.exp_path, 'karr.png'))

        # replicate mask to 3 color channels
        arr = np.zeros(karr.shape+(3,))
        for p in range(ch):
            arr[..., p] = karr

        # threshold values with insufficient difference
        mask = np.divide(arr, arr.max(), out=np.zeros_like(arr), where=arr != 0)
        th = 0.1
        mask[mask < th] = 0
        mask[mask >= th] = 1

        # ignore small regions
        mask = self.retain_connected(mask, n=4)

        # generate mask
        if self.cfg.params[self.cfg.opt_dbug]:
            full_mask = np.zeros(img.shape)
            if full_mask[j::2, ...].shape[0] == mask.shape[0]:
                full_mask[j::2, ...] = mask
            else:
                full_mask[j::2, ...][:mask.shape[0], ...] = mask
            misc.save_img_file(full_mask, os.path.join(self.cfg.exp_path, 'hex_filter_mask.png'), tag=True)

        # generate image indicating which pixels were treated
        if self.cfg.params[self.cfg.opt_dbug]:
            tes = img.copy()
            for p in range(ch):
                if tes[j::2, ...].shape[0] == mask.shape[0]:
                    tes[j::2, :, p][mask[..., p] != 0] = tes.max()
                else:
                    tes[j::2, :, p][:mask.shape[0], ...][mask[..., p] != 0] = tes.max()
            tes_out = np.divide(tes, tes.max(), out=np.zeros_like(tes), where=tes != 0)
            misc.save_img_file(tes_out, os.path.join(self.cfg.exp_path, 'ident.png'), tag=True)

        idxs = np.array(mask.nonzero())
        r = 1
        valid_idx = np.where((idxs[0] > r) & (idxs[1] > r) & (img[j::2].shape[0]-idxs[0] > r) & (img.shape[1]-idxs[1] > r))[0]
        idxs_vals = list(zip(idxs[0][valid_idx], idxs[1][valid_idx], idxs[2][valid_idx]))

        res = img.copy()
        for idx in idxs_vals:
            new_val = np.mean(img[i::2, ...][idx[0]:idx[0]+2, idx[1], idx[2]], axis=0)
            res[j::2, ...][idx[0], idx[1], idx[2]] = new_val

        if self.cfg.params[self.cfg.opt_dbug]:
            misc.save_img_file(res, os.path.join(self.cfg.exp_path, 'hex-corr.png'))

        return res

    @staticmethod
    def retain_connected(mask, n):

        w = mask.copy().T.ravel().astype('bool')

        idx = np.flatnonzero(np.r_[True, ~w, True])
        lens = np.diff(idx) - 1
        comb = np.vstack(list(zip(idx, lens)))
        filt = comb[comb[:, 1] >= n]
        vals = np.zeros_like(w)
        for f in filt:
            vals[f[0]:f[0]+f[1]] = 1

        return np.reshape(vals, mask.T.shape).T.astype('uint8')



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.lfp_aligner.lfp_microlenses import LfpMicroLenses

import numpy as np


class LfpCropper(LfpMicroLenses):

    def __init__(self, *args, **kwargs):
        super(LfpCropper, self).__init__(*args, **kwargs)

        # use _k as crop margin
        self._k = (self._limg_pitch - self._size_pitch).astype('int') // 2

        if self._lfp_img_align is not None:

            self._LENS_Y_MAX = int(self._lfp_img_align.shape[0]/self._limg_pitch[0])
            self._LENS_X_MAX = int(self._lfp_img_align.shape[1]/self._limg_pitch[1])
            p = self._lfp_img_align.shape[-1] if len(self._lfp_img_align.shape) == 3 else 1
            self.new_lfp_img = np.zeros([int(self._size_pitch*self._LENS_Y_MAX),
                                         int(self._size_pitch*self._LENS_X_MAX),
                                         p],
                                        dtype=self._lfp_img_align.dtype)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # reduce light field in angular domain (depending on settings)
        if self._size_pitch < self._limg_pitch[0] or self._size_pitch < self._limg_pitch[1]:
            self.proc_lens_iter(self.crop_micro_image, msg='Render angular domain')
        elif self._size_pitch == min(self._limg_pitch):
            self.new_lfp_img = self._lfp_img_align

    def crop_micro_image(self, ly, lx):

        self.new_lfp_img[ly*self._size_pitch:(ly+1)*self._size_pitch, lx*self._size_pitch:(lx+1)*self._size_pitch] = \
            self._lfp_img_align[self._k[0]+ly*self._limg_pitch[0]:(ly+1)*self._limg_pitch[0]-self._k[0],
                                self._k[1]+lx*self._limg_pitch[1]:(lx+1)*self._limg_pitch[1]-self._k[1]]

    @property
    def lfp_img_align(self):
        return self.new_lfp_img.copy() if self.new_lfp_img is not None else None



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.cfg import PlenopticamConfig
from plenopticam import misc

import numpy as np
import os


class LfpMicroLenses(object):

    def __init__(self, *args, **kwargs):

        # variables
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else PlenopticamConfig()
        self.sta = kwargs['sta'] if 'sta' in kwargs else misc.PlenopticamStatus()
        self._hex_odd = None
        self._lfp_img = kwargs['lfp_img'] if 'lfp_img' in kwargs else None
        self._wht_img = kwargs['wht_img'] if 'wht_img' in kwargs else None
        self._lfp_img_align = kwargs['lfp_img_align'] if 'lfp_img_align' in kwargs else None
        self._flip = kwargs['flip'] if 'flip' in kwargs else False

        # convert to float
        self._lfp_img = self._lfp_img.astype('float64') if self._lfp_img is not None else None
        self._wht_img = self._wht_img.astype('float64') if self._wht_img is not None else None

        if self.cfg.calibs:
            # micro lens array variables
            self._CENTROIDS = np.asarray(self.cfg.calibs[self.cfg.mic_list])
            self._LENS_Y_MAX = int(max(self._CENTROIDS[:, 2])+1)    # +1 to account for index 0
            self._LENS_X_MAX = int(max(self._CENTROIDS[:, 3])+1)    # +1 to account for index 0
            self._PAT_TYPE = self.cfg.calibs[self.cfg.pat_type] if self.cfg.pat_type in self.cfg.calibs else 'rec'
            self._hex_odd = self.get_hex_direction(self._CENTROIDS)

        # initialize micro image size and respective centers
        self._size_pitch = 0
        self._cent_pitch = 0

        # get pitch from aligned light field
        self._limg_pitch = self.lfp_align_pitch()

        # get mean pitch from centroids
        self._cavg_pitch = self.centroid_avg_pitch()

        # evaluate pitch sizes from aligned image, centroid list and user definition to obtain safe pitch
        self._size_pitch = self.safe_pitch_eval(
                                                cavg_pitch=min(self._cavg_pitch),
                                                limg_pitch=min(self._limg_pitch),
                                                user_pitch=int(self.cfg.params[self.cfg.ptc_leng]),
                                                )
        # micro image centers
        self._cent_pitch = self._size_pitch // 2

        try:
            self._DIMS = self._lfp_img.shape if len(self._lfp_img.shape) == 3 else self._lfp_img.shape + (1,)
        except (TypeError, AttributeError):
            pass
        except IndexError:
            self.sta.status_msg('Incompatible image dimensions: Please either use KxLx3 or KxLx1 array dimensions')
            self.sta.error = True

    def proc_lens_iter(self, fun, **kwargs):
        """ process light-field based on provided function handle and argument data """

        # status message handling
        msg = kwargs['msg'] if 'msg' in kwargs else 'Light-field alignment process'
        usr_prnt = kwargs['prnt'] if 'prnt' in kwargs else True
        if usr_prnt:
            self.sta.status_msg(msg, self.cfg.params[self.cfg.opt_prnt])

        args = [kwargs[key] for key in kwargs.keys() if key not in ('cfg', 'sta', 'msg', 'prnt')]

        try:
            # iterate through each micro lens centroid
            for ly in range(self._LENS_Y_MAX):
                for lx in range(self._LENS_X_MAX):

                    # perform provided function
                    fun(ly, lx, *args)

                # print progress status
                self.sta.progress((ly + 1) / self._LENS_Y_MAX * 100, self.cfg.params[self.cfg.opt_prnt])

                # check interrupt status
                if self.sta.interrupt:
                    return False

        except Exception as e:
            raise e

        return True

    def get_coords_by_idx(self, ly: int, lx: int) -> (float, float):
        """ yields micro image center in 2-D image coordinates """

        # filter mic by provided indices
        mic = self._CENTROIDS[(self._CENTROIDS[:, 2] == ly) & (self._CENTROIDS[:, 3] == lx), [0, 1]]

        return mic[0], mic[1]

    def safe_pitch_eval(self, cavg_pitch: float, limg_pitch: float, user_pitch: int) -> int:
        """ evaluate pitch size that is safe to use """

        safe_pitch = 3

        # comparison of user selected pitch and average centroid size
        if safe_pitch <= user_pitch <= cavg_pitch+2:  # allow user pitch to be slightly bigger than estimate
            safe_pitch = user_pitch
        elif user_pitch > cavg_pitch:
            safe_pitch = cavg_pitch
            msg_str = 'User size ({0} px) is larger than micro image size and reduced to {1} pixels.'
            self.sta.status_msg(msg_str.format(user_pitch, cavg_pitch), self.cfg.params[self.cfg.opt_prnt])
        elif user_pitch < safe_pitch < cavg_pitch:
            safe_pitch = cavg_pitch
            msg_str = 'User size ({0} px) is too small and increased to {1} pixels.'
            self.sta.status_msg(msg_str.format(user_pitch, cavg_pitch), self.cfg.params[self.cfg.opt_prnt])
        elif user_pitch < safe_pitch and cavg_pitch < safe_pitch:
            self.sta.status_msg('Micro image dimensions are too small for light field computation.', True)
            self.sta.interrupt = True

        # if chosen safe micro image size is larger than in the aligned image
        if 0 < limg_pitch < safe_pitch:
            # remove existing pickle file
            fp = os.path.join(self.cfg.exp_path, 'lfp_img_align.pkl')
            os.remove(fp)
            # status update
            self.sta.status_msg('Angular resolution mismatch in previous alignment. Redo process')
            self.sta.interrupt = True
        # check if micro image size in valid range
        elif limg_pitch >= safe_pitch > 0:
            self.cfg.params[self.cfg.ptc_leng] = safe_pitch
        # check if micro image size not set
        elif limg_pitch == 0:
            self._size_pitch = safe_pitch
            self.cfg.params[self.cfg.ptc_leng] = safe_pitch

        return int(safe_pitch)

    def centroid_avg_pitch(self) -> np.ndarray:
        """ estimate micro image pitch only from centroids """

        if not hasattr(self, '_CENTROIDS'):
            return np.zeros(2)

        # convert to numpy array
        centroids = np.asarray(self._CENTROIDS)

        # estimate maximum patch size
        central_row_idx = int(max(centroids[:, 2])/2)
        central_col_idx = int(max(centroids[:, 3])/2)
        mean_pitch_x = int(np.ceil(np.mean(np.diff(centroids[centroids[:, 2] == central_row_idx, 1]))))
        mean_pitch_y = int(np.ceil(np.mean(np.diff(centroids[centroids[:, 3] == central_col_idx, 0]))))

        return np.array([mean_pitch_y, mean_pitch_x])

    def centroid_align_pitch(self) -> int:
        """ obtain micro image pitch of aligned light-field from number of centroids """

        # estimate patch size
        lens_max_y = self._CENTROIDS[:][2].max() + 1     # +1 to account for index 0
        lens_max_x = self._CENTROIDS[:][3].max() + 1     # +1 to account for index 0
        pitch_estimate_y = self._lfp_img.shape[0]/lens_max_y
        pitch_estimate_x = self._lfp_img.shape[1]/lens_max_x

        if pitch_estimate_y-int(pitch_estimate_y) != 0 or pitch_estimate_x-int(pitch_estimate_x) != 0:
            msg = 'Micro image patch size error. Remove output folder or select re-calibration in settings.'
            self.sta.status_msg(msg=msg, opt=self.cfg.params[self.cfg.opt_prnt])
            self.sta.error = True

        return int(pitch_estimate_y)

    def lfp_align_pitch(self) -> np.ndarray:
        """ estimate pitch size from aligned light-field """

        # initialize output variable (return zero if aligned light field not present)
        self._limg_pitch = np.zeros(2, dtype='int')
        if self._lfp_img_align is None:
            return self._limg_pitch

        # obtain micro image size from both dimensions (horizontal accounts for hexagonal stretching)
        if hasattr(self, '_LENS_Y_MAX') and hasattr(self, '_LENS_X_MAX'):
            h = np.sqrt(3)/2 if self._PAT_TYPE == 'hex' else 1
            self._limg_pitch[0] = int(round(self._lfp_img_align.shape[0] / self._LENS_Y_MAX))
            self._limg_pitch[1] = int(round(self._lfp_img_align.shape[1] / self._LENS_X_MAX * h))
        else:
            for axis in [0, 1]:
                # take pitch where remainder of ratio between aligned image dimensions and candidate size is zero
                self._limg_pitch[axis] = self.remainder_ratio(num=self._lfp_img_align.shape[axis])

        return self._limg_pitch.astype('int')

    @staticmethod
    def remainder_ratio(num: int) -> int:
        # iterate through potential (odd or even) micro image size candidates
        for d in np.arange(3, 151, 1):
            if (num / d) % 1 == 0:
                return int(d)

        return 0

    @staticmethod
    def get_hex_direction(centroids: np.ndarray) -> bool:
        """ check if lower neighbor of upper left micro image center is shifted to left or right in hex grid

        :param centroids: phased array data
        :return: True if shifted to right
        """

        # get upper left MIC
        first_mic = centroids[(centroids[:, 2] == 0) & (centroids[:, 3] == 0), [0, 1]]

        # retrieve horizontal micro image shift (to determine search range borders)
        central_row_idx = int(centroids[:, 3].max()/2)
        mean_pitch = np.mean(np.diff(centroids[centroids[:, 3] == central_row_idx, 0]))

        # try to find MIC in lower left range (considering hexagonal order)
        found_mic = centroids[(centroids[:, 0] > first_mic[0]+mean_pitch/2) &
                              (centroids[:, 0] < first_mic[0]+3*mean_pitch/2) &
                              (centroids[:, 1] < first_mic[1]) &
                              (centroids[:, 1] > first_mic[1]-3*mean_pitch/4)].ravel()

        # true if MIC of next row lies on the right (false otherwise)
        hex_odd = True if found_mic.size == 0 else False

        return hex_odd

    @property
    def lfp_img(self):
        return self._lfp_img.copy() if self._lfp_img is not None else False

    @property
    def lfp_img_align(self):
        return self._lfp_img_align.copy() if self._lfp_img_align is not None else None



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


import numpy as np
from scipy.interpolate import interp2d
from color_space_converter import yuv_conv

from plenopticam.misc.normalizer import Normalizer


def create_gauss_kernel(length: int = 7, sigma: float = 1) -> np.ndarray:
    
    # ensure length is odd
    length = (length - 1)//2 + (length + 1)//2
    
    # compute Gaussian kernel
    xs = np.arange(-length//2 + 1, length//2 + 1)
    xx, yy = np.meshgrid(xs, xs)
    kernel = np.exp(-(xx**2 + yy**2) / (2*sigma**2))
    kernel /= kernel.sum()
    
    return kernel


def safe_get(any_dict, *keys):

    if any_dict:
        for key in keys:
            try:
                any_dict = any_dict[key]
            except KeyError:
                return None

    return any_dict


def img_resize(img, x_scale=1, y_scale=None, method=None, new_shape=None, norm_opt=False):
    """ perform image interpolation based on scipy lib """

    if not y_scale:
        y_scale = x_scale

    if x_scale == 1 and y_scale == 1 and new_shape is None or x_scale <= 0 or y_scale <= 0:
        return img

    method = 'cubic' if method is None else method
    dtype = img.dtype

    if len(img.shape) == 3:
        n, m, p = img.shape
    elif len(img.shape) == 2:
        n, m, p = img.shape + (1,)
        img = img[..., np.newaxis]
    else:
        raise NotImplementedError

    # construct new 2-D shape
    y_len, x_len = (int(round(n*y_scale)), int(round(m*x_scale))) if x_scale != 1 or y_scale != 1 else new_shape

    # interpolate
    new_img = np.zeros([y_len, x_len, p])
    for p in range(p):
        f = interp2d(range(m), range(n), img[:, :, p], kind=method)
        new_img[:, :, p] = f(np.linspace(0, m - 1, x_len), np.linspace(0, n - 1, y_len))

    # normalize to the 0-1 range
    new_img = Normalizer(new_img).type_norm(new_min=0, new_max=1) if norm_opt else new_img

    # remove added third axes in monochromatic image
    new_img = new_img[..., 0] if new_img.shape[-1] == 1 else new_img

    return new_img.astype(dtype)


def eq_channels(img):
    """ equalize channels of RGB image (make channels of even power) """

    chs = np.ones(img.shape[2]) if len(img.shape) == 3 else 1
    ch_max = np.argmax(img.sum(axis=0).sum(axis=0))
    for idx in range(len(chs)):
        chs[idx] = np.mean(img[..., ch_max]) / np.mean(img[..., idx])
        img[..., idx] *= chs[idx]

    return img


def robust_awb(img, t=0.3, max_iter=1000):
    """ inspired by Jun-yan Huo et al. and http://web.stanford.edu/~sujason/ColorBalancing/Code/robustAWB.m """

    img = Normalizer(img).type_norm(new_min=0, new_max=1.0)
    ref_pixel = img[0, 0, :].copy()

    u = .01  # gain step size
    a = .8  # double step threshold
    b = .001  # convergence threshold

    gains_adj = np.array([1., 1., 1.])

    for i in range(max_iter):
        img_yuv = yuv_conv(img)
        f = (abs(img_yuv[..., 1]) + abs(img_yuv[..., 2])) / img_yuv[..., 0]
        grays = np.zeros(img_yuv.shape)
        grays[f < t] = img_yuv[f < t]
        if np.sum(f < t) == 0:
            print('No valid gray pixels found.')
            break

        u_bar = np.mean(grays[..., 1])  # estimate
        v_bar = np.mean(grays[..., 2])  # estimate

        # rgb_est = yuv_conv(np.array([100, u_bar, v_bar]), inverse=True)    # convert average gray from YUV to RGB

        # U > V: blue needs adjustment otherwise red is treated
        err, ch = (u_bar, 2) if abs(u_bar) > abs(v_bar) else (v_bar, 0)

        if abs(err) >= a:
            delta = 2 * np.sign(err) * u  # accelerate gain adjustment if far off
        elif abs(err) < b:  # converged when u_bar and v_bar < b
            # delta = 0
            #self.sta.status_msg('AWB convergence reached', self.cfg.params[self.cfg.opt_prnt])
            break
        else:
            delta = err * u

        # negative feedback loop
        gains_adj[ch] -= delta

        img = np.dot(img, np.diag(gains_adj))

    # take gains only if result is obtained by convergence
    gains = img[0, 0, :] / ref_pixel if i != max_iter - 1 else (1, 1, 1)

    return img, gains


def suppress_user_warning(switch=None, category=None):

    import warnings
    switch = switch if switch is None else True
    if switch:
        warnings.filterwarnings("ignore", category=category)
    else:
        warnings.filterwarnings("default", category=category)



# local imports
from plenopticam.misc.type_checks import rint
from plenopticam.lfp_aligner.lfp_microlenses import LfpMicroLenses

# external libs
import numpy as np
import functools
from scipy.interpolate import interp2d, RectBivariateSpline


class LfpLocalResampler(LfpMicroLenses):

    def __init__(self, *args, **kwargs):
        super(LfpLocalResampler, self).__init__(*args, **kwargs)

        # interpolation method initialization
        method = kwargs['method'] if 'method' in kwargs else 'linear'
        method = method if method in ['nearest', 'linear', 'cubic', 'quintic'] else 'linear'
        method = 'cubic' if method == 'quintic' and self._size_pitch < 5 else method
        interp2d_method = functools.partial(interp2d, kind=method) if method is not None else interp2d

        if method is None:
            self._interpol_method = RectBivariateSpline
        elif method == 'nearest':
            self._interpol_method = self._nearest
        else:
            self._interpol_method = interp2d_method

    def local_resampling(self):
        """ cropping micro images to square shape while interpolating around their detected center (MIC) """

        # check interrupt status
        if self.sta.interrupt:
            return False

        # start resampling process (taking micro lens arrangement into account)
        if self.cfg.calibs[self.cfg.pat_type] == 'rec':
            self.resample_rec()
        elif self.cfg.calibs[self.cfg.pat_type] == 'hex':
            self.resample_hex()

        return True

    def _patch_align(self, window, mic):

        # initialize patch
        patch = np.zeros(window.shape)

        # verify patch shapes as wrong shapes cause crashes
        if window.shape[0] == self._size_pitch+2 and window.shape[1] == self._size_pitch+2:
            # iterate through color channels
            for p in range(self._DIMS[2]):

                fun = self._interpol_method(range(window.shape[1]), range(window.shape[0]), window[:, :, p])

                patch[:, :, p] = fun(np.arange(window.shape[1])+mic[1]-rint(mic[1]),
                                     np.arange(window.shape[0])+mic[0]-rint(mic[0]))
        else:
            self.sta.status_msg('Warning: chosen micro image size exceeds light-field borders')
            return np.zeros((self._size_pitch+2,)*2+(window.shape[2],))

        # flip patch to compensate for micro lens rotation
        patch = np.flip(patch, axis=(0, 1)) if self._flip else patch

        return patch

    def _nearest(self, range0, range1, window):

        def shift_win(shifted_range0, shifted_range1):
            range0 = np.round(shifted_range0).astype('int')
            range1 = np.round(shifted_range1).astype('int')
            return window[range0[0]:range0[-1]+1, range1[0]:range0[-1]+1]

        return shift_win

    def resample_rec(self):

        # initialize variables required for micro image resampling process
        self._lfp_img_align = np.zeros([self._LENS_Y_MAX * self._size_pitch,
                                        self._LENS_X_MAX * self._size_pitch,
                                        self._DIMS[2]])

        # iterate over each MIC
        for ly in range(self._LENS_Y_MAX):
            for lx in range(self._LENS_X_MAX):

                # find MIC by indices
                mic = self.get_coords_by_idx(ly=ly, lx=lx)

                # interpolate each micro image with its MIC as the center with consistent micro image size
                window = self._lfp_img[rint(mic[0]) - self._cent_pitch - 1:rint(mic[0]) + self._cent_pitch + 2,
                                       rint(mic[1]) - self._cent_pitch - 1:rint(mic[1]) + self._cent_pitch + 2]
                self._lfp_img_align[ly*self._size_pitch:(ly+1)*self._size_pitch,
                                    lx*self._size_pitch:(lx+1)*self._size_pitch] = \
                    self._patch_align(window, mic)[1:-1, 1:-1]

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status for on console
            self.sta.progress((ly + 1) / self._LENS_Y_MAX * 100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def resample_hex(self):

        # initialize variables required for micro image resampling process
        patch_stack = np.zeros([self._LENS_X_MAX, self._size_pitch, self._size_pitch, self._DIMS[2]])
        hex_stretch = int(np.round(2 * self._LENS_X_MAX / np.sqrt(3)))
        interp_stack = np.zeros([hex_stretch, self._size_pitch, self._size_pitch, self._DIMS[2]])
        self._lfp_img_align = np.zeros([self._LENS_Y_MAX*self._size_pitch, hex_stretch*self._size_pitch, self._DIMS[2]])

        # check if lower neighbor of upper left MIC is shifted to left or right
        hex_odd = self.get_hex_direction(self._CENTROIDS)

        # iterate over each MIC
        for ly in range(self._LENS_Y_MAX):
            for lx in range(self._LENS_X_MAX):

                # find MIC by indices
                mic = self.get_coords_by_idx(ly=ly, lx=lx)

                # interpolate each micro image with its MIC as the center and consistent micro image size
                window = self._lfp_img[rint(mic[0]) - self._cent_pitch - 1:rint(mic[0]) + self._cent_pitch + 2,
                         rint(mic[1]) - self._cent_pitch - 1:rint(mic[1]) + self._cent_pitch + 2]
                patch_stack[lx, :, :] = self._patch_align(window, mic)[1:-1, 1:-1]

            # image stretch interpolation in x-direction to compensate for hex-alignment
            for y in range(self._size_pitch):
                for x in range(self._size_pitch):
                    for p in range(self._DIMS[2]):
                        # stack of micro images elongated in x-direction
                        interp_coords = np.linspace(0, self._LENS_X_MAX, int(np.round(self._LENS_X_MAX*2/np.sqrt(3))))+\
                                        .5*np.mod(ly+hex_odd, 2)
                        interp_stack[:, y, x, p] = np.interp(interp_coords, range(self._LENS_X_MAX), patch_stack[:, y, x, p])

            self._lfp_img_align[ly*self._size_pitch:(ly+1)*self._size_pitch, :] = \
                np.concatenate(interp_stack, axis=1).reshape((self._size_pitch,
                                                              hex_stretch*self._size_pitch,
                                                              self._DIMS[2]))

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status
            self.sta.progress((ly+1) / self._LENS_Y_MAX * 100, self.cfg.params[self.cfg.opt_prnt])



from .lfp_microlenses import LfpMicroLenses
from .lfp_resampler import LfpResampler
from .lfp_rotator import LfpRotator
from .lfp_devignetter import LfpDevignetter
from .cfa_processor import CfaProcessor
from .cfa_outliers import CfaOutliers
from .top_level import LfpAligner
